@@ -6,14 +6,14 @@ import sys
 from datetime import datetime
 from logging import Logger
 from pypomes_core import TZ_LOCAL, exc_format
+from pypomes_crypto import jwt_get_claim, jwt_validate
 from typing import Any
 
 from .iam_common import (
-    IamServer, ServerParam, UserParam, _iam_lock,
-    _get_iam_users, _get_iam_registry, _get_public_key,
+    IamServer, ServerParam, UserParam, _iam_lock, _get_iam_users,
+    _get_iam_registry, _get_iam_property, _get_public_key,
     _get_login_timeout, _get_user_data, _iam_server_from_issuer
 )
-from .token_pomes import token_get_values, token_validate
 
 
 def iam_login(iam_server: IamServer,
@@ -23,7 +23,7 @@ def iam_login(iam_server: IamServer,
     """
     Build the URL for redirecting the request to *iam_server*'s authentication page.
 
-    These are the expected attributes in *args*:
+    The expected attributes in *args* are:
         - user-id: optional, identifies the reference user (alias: 'login')
         - redirect-uri: a parameter to be added to the query part of the returned URL
         -target-idp: optionally, identify a target identity provider for the login operation
@@ -56,7 +56,7 @@ def iam_login(iam_server: IamServer,
     # ('oauth_state' is a randomly-generated string, thus 'user_data' is always a new entry)
     oauth_state: str = "".join(secrets.choice(string.ascii_letters + string.digits) for _ in range(16))
     if target_idp:
-        oauth_state += f"#idp={target_idp}"
+        oauth_state += f"_{target_idp}"
 
     with _iam_lock:
         # retrieve the user data from the IAM server's registry
@@ -101,6 +101,7 @@ def iam_logout(iam_server: IamServer,
     Logout the user, by removing all data associating it from *iam_server*'s registry.
 
     The user is identified by the attribute *user-id* or *login*, provided in *args*.
+
     A logout request is sent to *iam_server* and, if successful, remove all data relating to the user
     from the *IAM* server's registry.
 
@@ -175,10 +176,75 @@ def iam_logout(iam_server: IamServer,
             errors.append(msg)
 
 
-def iam_get_token(iam_server: IamServer,
-                  args: dict[str, Any],
-                  errors: list[str] = None,
-                  logger: Logger = None) -> dict[str, str]:
+def iam_refresh(iam_server: IamServer,
+                args: dict[str, Any],
+                errors: list[str] = None,
+                logger: Logger = None) -> dict[str, str]:
+    """
+    Refresh the authentication token for the user, from *iam_server*.
+
+    The expected parameters in *args* are:
+        - user-id: identification for the reference user (alias: 'login')
+        - access-token: the old, probable expíred, token to be refreshed
+
+    On success, the returned *dict* will contain the following JSON:
+        {
+            "access-token": <token>,
+            "user-id": <user-identification
+        }
+
+    :param iam_server: the reference registered *IAM* server
+    :param args: the arguments passed when requesting the service
+    :param errors: incidental error messages
+    :param logger: optional logger
+    :return: the user identification and token issued, or *None* if error
+    """
+    # initialize the return variable
+    result: dict[str, str] | None = None
+
+    # obtain the user's identification ant the token
+    user_id: str = args.get("user-id") or args.get("login")
+    user_token: str = args.get("access-token")
+
+    err_msg: str | None = None
+    if user_id:
+        with _iam_lock:
+            # retrieve the user data in the IAM server's registry
+            user_data: dict[str, Any] = _get_user_data(iam_server=iam_server,
+                                                       user_id=user_id,
+                                                       errors=errors,
+                                                       logger=logger)
+            # retrieve the stored access token
+            access_token: str = user_data[UserParam.ACCESS_TOKEN] if user_data else None
+            if access_token:
+                if access_token == user_token:
+                    result = __retrieve_token(iam_server=iam_server,
+                                              user_id=user_id,
+                                              errors=errors,
+                                              logger=logger)
+                else:
+                    err_msg = "Tokens do not match"
+                    if logger:
+                        logger.error(msg=err_msg)
+            else:
+                err_msg = f"User '{user_id}' not authenticated"
+                if logger:
+                    logger.error(msg=err_msg)
+    else:
+        err_msg = "User identification not provided"
+        if logger:
+            logger.error(msg=err_msg)
+
+    if err_msg and isinstance(errors, list):
+        errors.append(err_msg)
+
+    return result
+
+
+def iam_token(iam_server: IamServer,
+              args: dict[str, Any],
+              errors: list[str] = None,
+              logger: Logger = None) -> dict[str, str]:
     """
     Retrieve the authentication token for the user, from *iam_server*.
 
@@ -199,81 +265,21 @@ def iam_get_token(iam_server: IamServer,
     # initialize the return variable
     result: dict[str, str] | None = None
 
-    # obtain the user's identification
+    # obtain the user's identification ant the token
     user_id: str = args.get("user-id") or args.get("login")
 
-    err_msg: str | None = None
     if user_id:
         with _iam_lock:
-            # retrieve the user data in the IAM server's registry
-            user_data: dict[str, Any] = _get_user_data(iam_server=iam_server,
-                                                       user_id=user_id,
-                                                       errors=errors,
-                                                       logger=logger)
-            # retrieve the stored access token
-            token: str = user_data[UserParam.ACCESS_TOKEN] if user_data else None
-            if token:
-                access_expiration: int = user_data.get(UserParam.ACCESS_EXPIRATION)
-                now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
-                if now < access_expiration:
-                    result = {
-                        "access-token": token,
-                        "user-id": user_id
-                    }
-                else:
-                    # access token has expired
-                    refresh_token: str = user_data[UserParam.REFRESH_TOKEN]
-                    if refresh_token:
-                        refresh_expiration: int = user_data[UserParam.REFRESH_EXPIRATION]
-                        if now < refresh_expiration:
-                            header_data: dict[str, str] = {
-                                "Content-Type": "application/x-www-form-urlencoded"
-                            }
-                            body_data: dict[str, str] = {
-                                "grant_type": "refresh_token",
-                                "refresh_token": refresh_token
-                            }
-                            now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
-                            token_data: dict[str, Any] = __post_for_token(iam_server=iam_server,
-                                                                          header_data=header_data,
-                                                                          body_data=body_data,
-                                                                          errors=errors,
-                                                                          logger=logger)
-                            # validate and store the token data
-                            if token_data:
-                                token_info: tuple[str, str] = __validate_and_store(iam_server=iam_server,
-                                                                                   user_data=user_data,
-                                                                                   token_data=token_data,
-                                                                                   now=now,
-                                                                                   errors=errors,
-                                                                                   logger=logger)
-                                result = {
-                                    "access-token": token_info[1],
-                                    "user-id": user_id
-                                }
-                            else:
-                                # refresh token is no longer valid
-                                user_data[UserParam.REFRESH_TOKEN] = None
-                        else:
-                            # refresh token has expired
-                            err_msg = "Access and refresh tokens expired"
-                            if logger:
-                                logger.error(msg=err_msg)
-                    else:
-                        err_msg = "Access token expired, no refresh token available"
-                        if logger:
-                            logger.error(msg=err_msg)
-            else:
-                err_msg = f"User '{user_id}' not authenticated"
-                if logger:
-                    logger.error(msg=err_msg)
+            result = __retrieve_token(iam_server=iam_server,
+                                      user_id=user_id,
+                                      errors=errors,
+                                      logger=logger)
     else:
-        err_msg = "User identification not provided"
+        msg: str = "User identification not provided"
         if logger:
-            logger.error(msg=err_msg)
-
-    if err_msg and isinstance(errors, list):
-        errors.append(err_msg)
+            logger.error(msg=msg)
+        if isinstance(errors, list):
+            errors.append(msg)
 
     return result
 
@@ -285,7 +291,7 @@ def iam_callback(iam_server: IamServer,
     """
     Entry point for the callback from *iam_server* via the front-end application, on authentication operations.
 
-    The relevant expected arguments in *args* are:
+    The expected arguments in *args* are:
         - *state*: used to enhance security during the authorization process, typically to provide *CSRF* protection
         - *code*: the temporary authorization code provided by *iam_server*, to be exchanged for the token
 
@@ -322,8 +328,8 @@ def iam_callback(iam_server: IamServer,
             if int(datetime.now(tz=TZ_LOCAL).timestamp()) > expiration:
                 errors.append("Operation timeout")
             else:
-                pos: int = oauth_state.rfind("#idp=")
-                target_idp: str = oauth_state[pos+4:] if pos > 0 else None
+                pos: int = oauth_state.rfind("_")
+                target_idp: str = oauth_state[pos+1:] if pos > 0 else None
                 target_iam: IamServer = IamServer(target_idp) if target_idp in IamServer else None
                 target_data: dict[str, Any] = user_data.copy() if target_iam else None
                 users.pop(oauth_state)
@@ -416,65 +422,71 @@ def iam_exchange(iam_server: IamServer,
     # initialize the return variable
     result: tuple[str, str] | None = None
 
+    # make sure to have an errors list
+    if not isinstance(errors, list):
+        errors = []
+
     # obtain the user's identification
     user_id: str = args.get("user-id") or args.get("login")
 
     # obtain the token to be exchanged
-    token: str = args.get("access-token") if user_id else None
-    token_issuer: tuple[str] = token_get_values(token=token,
-                                                keys=("iss",),
-                                                errors=errors,
-                                                logger=logger)
-    if not errors:
-        with _iam_lock:
-            # retrieve the IAM server's registry
-            registry: dict[str, Any] = _get_iam_registry(iam_server=iam_server,
-                                                         errors=errors,
-                                                         logger=logger)
-            if registry:
-                # make sure 'client_id' is linked to the token's 'token_sub' at the IAM server
-                __assert_link(iam_server=iam_server,
-                              user_id=user_id,
-                              token=token,
-                              token_issuer=token_issuer[0],
-                              errors=errors,
-                              logger=logger)
-                if not errors:
-                    # exchange the token
-                    if logger:
-                        logger.debug(msg=f"Requesting the token exchange to IAM server '{iam_server}'")
-                    header_data: dict[str, Any] = {
-                        "Content-Type": "application/x-www-form-urlencoded"
-                    }
-                    body_data: dict[str, str] = {
-                        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
-                        "subject_token": token,
-                        "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
-                        "requested_token_type": "urn:ietf:params:oauth:token-type:access_token",
-                        "audience": registry[ServerParam.CLIENT_ID],
-                        "subject_issuer": token_issuer
-                    }
-                    now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
-                    token_data: dict[str, Any] = __post_for_token(iam_server=iam_server,
-                                                                  header_data=header_data,
-                                                                  body_data=body_data,
-                                                                  errors=errors,
-                                                                  logger=logger)
-                    # validate and store the token data
-                    if token_data:
-                        user_data: dict[str, Any] = {}
-                        result = __validate_and_store(iam_server=iam_server,
-                                                      user_data=user_data,
-                                                      token_data=token_data,
-                                                      now=now,
-                                                      errors=errors,
-                                                      logger=logger)
+    token: str = args.get("access-token")
+
+    if user_id and token:
+        token_issuer: str = jwt_get_claim(token=token,
+                                          key="iss",
+                                          errors=errors,
+                                          logger=logger)
+        if not errors:
+            with _iam_lock:
+                # retrieve the IAM server's registry
+                client_id: str = _get_iam_property(iam_server=iam_server,
+                                                   attr=ServerParam.CLIENT_ID,
+                                                   errors=errors,
+                                                   logger=logger)
+                if client_id:
+                    # make sure 'client_id' is linked to the token's 'token_sub' at the IAM server
+                    __assert_link(iam_server=iam_server,
+                                  user_id=user_id,
+                                  token=token,
+                                  token_issuer=token_issuer,
+                                  errors=errors,
+                                  logger=logger)
+                    if not errors:
+                        # exchange the token
+                        if logger:
+                            logger.debug(msg=f"Requesting the token exchange to IAM server '{iam_server}'")
+                        header_data: dict[str, Any] = {
+                            "Content-Type": "application/x-www-form-urlencoded"
+                        }
+                        body_data: dict[str, str] = {
+                            "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
+                            "subject_token": token,
+                            "subject_token_type": "urn:ietf:params:oauth:token-type:access_token",
+                            "requested_token_type": "urn:ietf:params:oauth:token-type:access_token",
+                            "audience": client_id,
+                            "subject_issuer": token_issuer
+                        }
+                        now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
+                        token_data: dict[str, Any] = __post_for_token(iam_server=iam_server,
+                                                                      header_data=header_data,
+                                                                      body_data=body_data,
+                                                                      errors=errors,
+                                                                      logger=logger)
+                        # validate and store the token data
+                        if not errors:
+                            user_data: dict[str, Any] = {}
+                            result = __validate_and_store(iam_server=iam_server,
+                                                          user_data=user_data,
+                                                          token_data=token_data,
+                                                          now=now,
+                                                          errors=errors,
+                                                          logger=logger)
     else:
         msg: str = "User identification or token not provided"
         if logger:
-            logger.error(msg=msg)
-        if isinstance(errors, list):
-            errors.append(msg)
+            logger.debug(msg=msg)
+        errors.append(msg)
 
     return result
 
@@ -550,6 +562,10 @@ def __assert_link(iam_server: IamServer,
     :param errors: incidental errors
     :param logger: optional logger
     """
+    # make sure to have an errors list
+    if not isinstance(errors, list):
+        errors = []
+
     if logger:
         logger.debug(msg="Verifying associations for user "
                          f"'{user_id}' in IAM server '{iam_server}'")
@@ -579,7 +595,7 @@ def __assert_link(iam_server: IamServer,
                                                      params=params,
                                                      errors=errors,
                                                      logger=logger)
-        if users:
+        if not errors:
             # verify whether the IAM server that issued the token is a federated identity provider
             # in the associations between 'user_id' and the internal user identification
             internal_id: str = users[0].get("id")
@@ -597,18 +613,18 @@ def __assert_link(iam_server: IamServer,
             provider_name: str = _iam_server_from_issuer(issuer=token_issuer,
                                                          errors=errors,
                                                          logger=logger)
-            if provider_name:
+            if not errors:
                 for provider in providers:
                     if provider.get("identityProvider") == provider_name:
                         no_link = False
                         break
                 if no_link:
                     # link the identities
-                    token_sub: tuple[str] = token_get_values(token=token,
-                                                             keys=("sub",),
-                                                             errors=errors,
-                                                             logger=logger)
-                    if token_sub:
+                    token_sub: str = jwt_get_claim(token=token,
+                                                   key="sub",
+                                                   errors=errors,
+                                                   logger=logger)
+                    if not errors:
                         if logger:
                             logger.debug(msg="Creating an association between identifications "
                                              f"'{user_id}' and '{token_sub}' in IAM server '{iam_server}'")
@@ -973,6 +989,92 @@ def __post_for_token(iam_server: IamServer,
     return result
 
 
+def __retrieve_token(iam_server: IamServer,
+                     user_id: str,
+                     errors: list[str] | None,
+                     logger: Logger | None) -> dict[str, str]:
+    """
+    Retrieve the token associated with *user_id* from the *IAM* registry.
+
+    :param iam_server: the reference registered *IAM* server
+    :param user_id: the identification for the user
+    :param errors: incidental errors
+    :param logger: optional logger
+    :return: the token data, or *None* if error
+    """
+    # initialize the return variable
+    result: dict[str, str] | None = None
+
+    # retrieve the user data in the IAM server's registry
+    user_data: dict[str, Any] = _get_user_data(iam_server=iam_server,
+                                               user_id=user_id,
+                                               errors=errors,
+                                               logger=logger)
+    # retrieve the stored access token
+    access_token: str = user_data[UserParam.ACCESS_TOKEN] if user_data else None
+    if access_token:
+        # access token has expired
+        refresh_token: str = user_data[UserParam.REFRESH_TOKEN]
+        if refresh_token:
+            access_expiration: int = user_data.get(UserParam.ACCESS_EXPIRATION)
+            now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
+            if now < access_expiration:
+                # token has not expired, so return it
+                result = {
+                    "access-token": access_token,
+                    "user-id": user_id
+                }
+            else:
+                # access token has expired
+                refresh_token: str = user_data[UserParam.REFRESH_TOKEN]
+                if refresh_token:
+                    refresh_expiration: int = user_data[UserParam.REFRESH_EXPIRATION]
+                    if now < refresh_expiration:
+                        header_data: dict[str, str] = {
+                            "Content-Type": "application/x-www-form-urlencoded"
+                        }
+                        body_data: dict[str, str] = {
+                            "grant_type": "refresh_token",
+                            "refresh_token": refresh_token
+                        }
+                        now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
+                        token_data: dict[str, Any] = __post_for_token(iam_server=iam_server,
+                                                                      header_data=header_data,
+                                                                      body_data=body_data,
+                                                                      errors=errors,
+                                                                      logger=logger)
+                        # validate and store the token data
+                        if token_data:
+                            token_info: tuple[str, str] = __validate_and_store(iam_server=iam_server,
+                                                                               user_data=user_data,
+                                                                               token_data=token_data,
+                                                                               now=now,
+                                                                               errors=errors,
+                                                                               logger=logger)
+                            result = {
+                                "access-token": token_info[1],
+                                "user-id": user_id
+                            }
+                        else:
+                            # refresh token is no longer valid
+                            user_data[UserParam.REFRESH_TOKEN] = None
+                    else:
+                        # refresh token has expired
+                        err_msg = "Access and refresh tokens expired"
+                        if logger:
+                            logger.error(msg=err_msg)
+                else:
+                    err_msg = "Access token expired, no refresh token available"
+                    if logger:
+                        logger.error(msg=err_msg)
+    else:
+        err_msg = f"User '{user_id}' not authenticated"
+        if logger:
+            logger.error(msg=err_msg)
+
+    return result
+
+
 def __validate_and_store(iam_server: IamServer,
                          user_data: dict[str, Any],
                          token_data: dict[str, Any],
@@ -1023,13 +1125,13 @@ def __validate_and_store(iam_server: IamServer,
             recipient_attr = registry[ServerParam.RECIPIENT_ATTR]
             login_id = user_data.pop("login-id", None)
             base_url: str = f"{registry[ServerParam.URL_BASE]}/realms/{registry[ServerParam.CLIENT_REALM]}"
-            claims: dict[str, dict[str, Any]] = token_validate(token=token,
-                                                               issuer=base_url,
-                                                               recipient_id=login_id,
-                                                               recipient_attr=recipient_attr,
-                                                               public_key=public_key,
-                                                               errors=errors,
-                                                               logger=logger)
+            claims: dict[str, dict[str, Any]] = jwt_validate(token=token,
+                                                             issuer=base_url,
+                                                             recipient_id=login_id,
+                                                             recipient_attr=recipient_attr,
+                                                             public_key=public_key,
+                                                             errors=errors,
+                                                             logger=logger)
             if claims:
                 users: dict[str, dict[str, Any]] = _get_iam_users(iam_server=iam_server,
                                                                   errors=errors,
