@@ -1,13 +1,12 @@
-import requests
 import sys
 from datetime import datetime
 from enum import StrEnum
 from logging import Logger
 from pypomes_core import (
-    APP_PREFIX, TZ_LOCAL, exc_format,
+    APP_PREFIX, TZ_LOCAL,
     env_get_int, env_get_str, env_get_strs
 )
-from pypomes_crypto import crypto_jwk_convert
+from pypomes_crypto import jwt_get_public_key
 from threading import RLock
 from typing import Any, Final
 
@@ -31,12 +30,14 @@ class ServerParam(StrEnum):
     ENDPOINT_CALLBACK_EXCHANGE = "endpoint-callback-exchange"
     ENDPOINT_LOGIN = "endpoint-login"
     ENDPOINT_LOGOUT = "endpoint_logout"
+    ENDPOINT_REFRESH = "endpoint-refresh"
     ENDPOINT_TOKEN = "endpoint-token"
     ENDPOINT_EXCHANGE = "endpoint-exchange"
     LOGIN_TIMEOUT = "login-timeout"
     PK_EXPIRATION = "pk-expiration"
     PK_LIFETIME = "pk-lifetime"
     RECIPIENT_ATTR = "recipient-attr"
+    TRUSTED_HOSTS = "trusted-hosts"
     # dynamic attributes
     PUBLIC_KEY = "public-key"
     URL_BASE = "url-base"
@@ -76,20 +77,28 @@ def __get_iam_data() -> dict[IamServer, dict[ServerParam, Any]]:
           - *<APP_PREFIX>_<IAM>_LOGIN_TIMEOUT*      (optional, defaults to no timeout)
           - *<APP_PREFIX>_<IAM>_PK_LIFETIME*        (optional, defaults to non-terminating lifetime)
           - *<APP_PREFIX>_<IAM>_RECIPIENT_ATTR*     (required)
+          - *<APP_PREFIX>_<IAM>_TRUSTED_HOSTS*      (optional)
           - *<APP_PREFIX>_<IAM>_URL_BASE*           (required)
 
     2. A group of special environment variables identifying endpoints for authentication services may be specified,
-       following the same scheme as presented in item *1* above. These are not part of the *IAM* server's setup,
-       but are meant to be used by function *iam_setup_endpoints()*, wherein the values in those variables
+       following the same scheme as presented in item *1* above. These are the second part of the *IAM* server's
+       setup, and are meant to be used by function *iam_setup_endpoints()*, wherein the values in those variables
        would represent default values for its parameters, respectively:
           - *<APP_PREFIX>_<IAM>_ENDPOINT_CALLBACK*
           - *<APP_PREFIX>_<IAM>_ENDPOINT_CALLBACK_EXCHANGE*
           - *<APP_PREFIX>_<IAM>_ENDPOINT_EXCHANGE*
           - *<APP_PREFIX>_<IAM>_ENDPOINT_LOGIN*
           - *<APP_PREFIX>_<IAM>_ENDPOINT_LOGOUT*
+          - *<APP_PREFIX>_<IAM>_ENDPOINT_REFRESH*
           - *<APP_PREFIX>_<IAM>_ENDPOINT_TOKEN*
           - *<APP_PREFIX>_<IAM>_ENDPOINT_USERINFO*
 
+    3. One of the above endpoints, namely *<APP_PREFIX>_<IAM>_ENDPOINT_TOKEN*, requires special protection.
+       By its very nature, it is expected to be seldomly needed, and if indeed used, limited in scope and
+       restricted to specific parties. This is the purpose of the special environment variable
+       *<APP_PREFIX>_<IAM>_TRUSTED_HOSTS*, which, if specified, will list the only requesting hosts
+       allowed to be serviced at that endpoint.
+
     :return: the configuration data for the select *IAM* servers.
     """
     # initialize the return variable
@@ -106,6 +115,7 @@ def __get_iam_data() -> dict[IamServer, dict[ServerParam, Any]]:
             ServerParam.LOGIN_TIMEOUT: env_get_str(key=f"{APP_PREFIX}_{prefix}_LOGIN_TIMEOUT"),
             ServerParam.PK_LIFETIME: env_get_int(key=f"{APP_PREFIX}_{prefix}_PK_LIFETIME"),
             ServerParam.RECIPIENT_ATTR: env_get_str(key=f"{APP_PREFIX}_{prefix}_RECIPIENT_ATTR"),
+            ServerParam.TRUSTED_HOSTS: env_get_strs(key=f"{APP_PREFIX}_{prefix}_TRUSTED_HOSTS"),
             ServerParam.URL_BASE: env_get_str(key=f"{APP_PREFIX}_{prefix}_URL_BASE"),
             # dynamically set
             ServerParam.PK_EXPIRATION: 0,
@@ -131,6 +141,7 @@ def __get_iam_data() -> dict[IamServer, dict[ServerParam, Any]]:
 #       "public-key": <str>,
 #       "pk-lifetime": <int>,
 #       "pk-expiration": <int>,
+#       "truted-requesters>: <list[str]>,
 #       "users": {}
 #    },
 #    ...
@@ -203,7 +214,7 @@ def _iam_server_from_issuer(issuer: str,
     for iam_server, registry in _IAM_SERVERS.items():
         base_url: str = f"{registry[ServerParam.URL_BASE]}/realms/{registry[ServerParam.CLIENT_REALM]}"
         if base_url == issuer:
-            result = type(IamServer)(iam_server)
+            result = IamServer(iam_server)
             break
 
     if not result:
@@ -222,34 +233,7 @@ def _get_public_key(iam_server: IamServer,
     """
     Obtain the public key used by *iam_server* to sign the authentication tokens.
 
-    This is accomplished by requesting the token issuer for its *JWKS* (JSON Web Key Set),
-    containing the public keys used for various purposes, as indicated in the attribute *use*:
-        - *enc*: the key is intended for encryption
-        - *sig*: the key is intended for digital signature
-        - *wrap*: the key is intended for key wrapping
-
-    A typical JWKS set has the following format (for simplicity, 'n' and 'x5c' are truncated):
-        {
-            "keys": [
-                {
-                    "kid": "X2QEcSQ4Tg2M2EK6s2nhRHZH_GwD_zxZtiWVwP4S0tg",
-                    "kty": "RSA",
-                    "alg": "RSA256",
-                    "use": "sig",
-                    "n": "tQmDmyM3tMFt5FMVMbqbQYpaDPf6A5l4e_kTVDBiHrK_bRlGfkk8hYm5SNzNzCZ...",
-                    "e": "AQAB",
-                    "x5c": [
-                        "MIIClzCCAX8CBgGZY0bqrTANBgkqhkiG9w0BAQsFADAPMQ0wCwYDVQQDDARpanVk..."
-                    ],
-                    "x5t": "MHfVp4kBjEZuYOtiaaGsfLCL15Q",
-                    "x5t#S256": "QADezSLgD8emuonBz8hn8ghTnxo7AHX4NVNkr4luEhk"
-                },
-                ...
-            ]
-        }
-
-    Once the signature key is obtained, it is converted from its original *JWK* (JSON Web Key) format
-    to *PEM* (Privacy-Enhanced Mail) format. The public key is saved in *iam_server*'s registry.
+    The signaature is obtained and stored in *PEM* (Privacy-Enhanced Mail) format.
 
     :param iam_server: the reference registered *IAM* server
     :param errors: incidental error messages
@@ -265,57 +249,17 @@ def _get_public_key(iam_server: IamServer,
     if registry:
         now: int = int(datetime.now(tz=TZ_LOCAL).timestamp())
         if now > registry[ServerParam.PK_EXPIRATION]:
-            # obtain the JWKS (JSON Web Key Set) from the token issuer
-            base_url: str = f"{registry[ServerParam.URL_BASE]}/realms/{registry[ServerParam.CLIENT_REALM]}"
-            url: str = f"{base_url}/protocol/openid-connect/certs"
-            if logger:
-                logger.debug(msg=f"Obtaining signature public key used by IAM server '{iam_server}'")
-                logger.debug(msg=f"GET {url}")
-            try:
-                response: requests.Response = requests.get(url=url)
-                if response.status_code == 200:
-                    # request succeeded
-                    if logger:
-                        logger.debug(msg=f"GET success, status {response.status_code}")
-                    # select the appropriate JWK
-                    reply: dict[str, list[dict[str, str]]] = response.json()
-                    jwk: dict[str, str] | None = None
-                    for key in reply["keys"]:
-                        if key.get("use") == "sig":
-                            jwk = key
-                            break
-                    if jwk:
-                        # convert from 'JWK' to 'PEM' and save it for further use
-                        result = crypto_jwk_convert(jwk=jwk,
-                                                    fmt="PEM")
-                        registry[ServerParam.PUBLIC_KEY] = result
-                        lifetime: int = registry[ServerParam.PK_LIFETIME] or 0
-                        registry[ServerParam.PK_EXPIRATION] = now + lifetime if lifetime else sys.maxsize
-                        if logger:
-                            logger.debug("Public key obtained and saved")
-                    else:
-                        msg = "Signature public key missing from the token issuer's JWKS"
-                        if logger:
-                            logger.error(msg=msg)
-                        if isinstance(errors, list):
-                            errors.append(msg)
-                elif logger:
-                    msg: str = f"GET failure, status {response.status_code}, reason {response.reason}"
-                    if hasattr(response, "content") and response.content:
-                        msg += f", content {response.content}"
-                    logger.error(msg=msg)
-                    if isinstance(errors, list):
-                        errors.append(msg)
-            except Exception as e:
-                # the operation raised an exception
-                msg = exc_format(exc=e,
-                                 exc_info=sys.exc_info())
-                if logger:
-                    logger.error(msg=msg)
-                if isinstance(errors, list):
-                    errors.append(msg)
-        else:
-            result = registry[ServerParam.PUBLIC_KEY]
+            # obtain the public key from the token issuer
+            issuer: str = f"{registry[ServerParam.URL_BASE]}/realms/{registry[ServerParam.CLIENT_REALM]}"
+            registry[ServerParam.PUBLIC_KEY] = jwt_get_public_key(issuer=issuer,
+                                                                  fmt="PEM",
+                                                                  errors=errors,
+                                                                  logger=logger)
+            lifetime: int = registry[ServerParam.PK_LIFETIME] or 0
+            registry[ServerParam.PK_EXPIRATION] = now + lifetime if lifetime else sys.maxsize
+
+    if not errors:
+        result = registry[ServerParam.PUBLIC_KEY]
 
     return result
 
@@ -408,6 +352,25 @@ def _get_iam_registry(iam_server: IamServer,
     return result
 
 
+def _get_iam_property(iam_server: IamServer,
+                      attr: ServerParam,
+                      errors: list[str] | None,
+                      logger: Logger | None) -> list[str] | str | int:
+    """
+    Retrieve the value of *attr* in the registry associated with *iam_server*.
+
+    :param iam_server: the reference registered *IAM* server
+    :param errors: incidental error messages
+    :param logger: optional logger
+    :return: the registry associated with *iam_server*, or *None* if the server is unknown
+    """
+    # obtain the registry
+    registry: dict[str, Any] = _get_iam_registry(iam_server=iam_server,
+                                                 errors=errors,
+                                                 logger=logger)
+    return registry[attr] if registry else None
+
+
 def _get_iam_users(iam_server: IamServer,
                    errors: list[str] | None,
                    logger: Logger | None) -> dict[str, dict[str, Any]]:
